#include <lib94/lib94.hpp>
#include <functional>
#include <algorithm>
#include <iostream>
#include <cassert>
#include <random>
#include <mpi.h>

namespace lib94 {

  static std::mt19937_64 prng;
  instruction core[LIB94_CORE_SIZE];

  void seed_prng(uint_fast64_t seed) {
    prng.seed(seed);
  }

  void clear_core(const instruction &background) {
    for (instruction &i : core)
      i = background;
  }

  void clear_core_random() {
    std::uniform_int_distribution<uint8_t> op(0, 15);
    std::uniform_int_distribution<uint8_t> mod(0, 6);
    std::uniform_int_distribution<uint8_t> modes(0, 7);
    std::uniform_int_distribution<number_t> number(0, LIB94_CORE_SIZE - 1);

    for (instruction &i : core)
      i = {
        .op = (opcode)op(prng),
        .mod = (modifier)mod(prng),
        .amode = (mode)modes(prng),
        .bmode = (mode)modes(prng),
        .anumber = number(prng),
        .bnumber = number(prng)
      };
  }

  static std::set<number_t> written_addresses;
  static std::set<number_t> read_addresses;
  static std::set<number_t> executed_addresses;

  void add_written_instruction(const instruction *instr) {
    written_addresses.insert(instr - core);
  }

  void add_read_instruction(const instruction *instr) {
    read_addresses.insert(instr - core);
  }

  struct warrior_info {
    const warrior *the_warrior;
    std::deque<number_t> processes;
  };

  static std::vector<warrior_info> warrior_infos;
  std::deque<warrior_info *> alive_warriors;

  bool init_round(const warrior *const *warriors, size_t count, const number_t *offsets, bool shuffle) {
    clear_address_sets();
    warrior_infos.clear();
    alive_warriors = std::deque<warrior_info *>();

    std::vector<number_t> offsets_vector;

    if (!offsets) {

      offsets_vector.resize(count);
      offsets_vector[0] = 0;

      std::uniform_real_distribution phi_dist(0.0, 1.0);
      number_t gap_remaining = LIB94_CORE_SIZE;

      for (size_t i = 0; i < count; ++i) {
        number_t wlength = warriors[i]->instructions.size();
        if (wlength > gap_remaining)
          return false;
        gap_remaining -= wlength;
      }

      for (size_t i = 1; i < count; ++i) {
        number_t gap_size = std::floor(gap_remaining * (1.0 - std::pow(phi_dist(prng), 1.0 / (count - i))));
        offsets_vector[i] = offsets_vector[i - 1] + warriors[i - 1]->instructions.size() + gap_size;
        gap_remaining -= gap_size;
      }

      offsets = offsets_vector.data();

    }

    for (size_t i = 0; i < count; ++i) {

      const warrior *w = warriors[i];
      const number_t offset = offsets[i];

      for (number_t i = 0; i < (number_t)w->instructions.size(); ++i) {
        assert(offset + i < LIB94_CORE_SIZE);
        core[offset + i] = w->instructions[i];
        add_written_instruction(core + offset + i);
      }

      warrior_infos.push_back({});
      warrior_info *wi = &warrior_infos.back();
      wi->the_warrior = w;
      assert(offset + w->org < LIB94_CORE_SIZE);
      wi->processes.push_back(offsets[i] + w->org);

    }

    if (shuffle)
      std::shuffle(warrior_infos.begin(), warrior_infos.end(), prng);

    for (warrior_info &wi : warrior_infos)
      alive_warriors.push_back(&wi);
    return true;
  }

  void remove_all_warriors() {
    alive_warriors.clear();
  }

  size_t alive_warrior_count() {
    return alive_warriors.size();
  }

  const warrior *get_next_warrior() {
    assert(alive_warriors.size() > 0);
    return alive_warriors.front()->the_warrior;
  }

  const std::deque<number_t> &get_processes(const warrior *for_warrior) {
    for (const warrior_info &wi : warrior_infos)
      if (wi.the_warrior == for_warrior)
        return wi.processes;
    assert(false);
  }

  const std::set<number_t> &get_written_addresses() {
    return written_addresses;
  }

  const std::set<number_t> &get_read_addresses() {
    return read_addresses;
  }

  const std::set<number_t> &get_executed_addresses() {
    return executed_addresses;
  }

  void clear_address_sets() {
    written_addresses.clear();
    read_addresses.clear();
    executed_addresses.clear();
  }

  const instruction &get_instruction(number_t address) {
    return core[address];
  }

  number_t program_counter;
  static instruction instruction_register;

  static number_t a_pointer, b_pointer;
  instruction a_instruction, b_instruction;

  instruction *a_instruction_writable;
  instruction *b_instruction_writable;

  static void evaluate_operand(mode m, number_t n, number_t &ptr, instruction &instr) {
    instruction *secondary;

    switch (m) {

      case IMMEDIATE:
        ptr = 0;
        instr = core[program_counter];
        break;

      case DIRECT:
        ptr = n;
        instr = core[(program_counter + n) % LIB94_CORE_SIZE];
        break;

      case A_INDIRECT:
        secondary = core + (program_counter + n) % LIB94_CORE_SIZE;
        ptr = (n + secondary->anumber) % LIB94_CORE_SIZE;
        instr = core[(program_counter + ptr) % LIB94_CORE_SIZE];
        add_read_instruction(secondary);
        break;

      case B_INDIRECT:
        secondary = core + (program_counter + n) % LIB94_CORE_SIZE;
        ptr = (n + secondary->bnumber) % LIB94_CORE_SIZE;
        instr = core[(program_counter + ptr) % LIB94_CORE_SIZE];
        add_read_instruction(secondary);
        break;

      case A_DECREMENT:
        secondary = core + (program_counter + n) % LIB94_CORE_SIZE;
        secondary->anumber = (secondary->anumber + LIB94_CORE_SIZE - 1) % LIB94_CORE_SIZE;
        ptr = (n + secondary->anumber) % LIB94_CORE_SIZE;
        instr = core[(program_counter + ptr) % LIB94_CORE_SIZE];
        add_read_instruction(secondary);
        add_written_instruction(secondary);
        break;

      case B_DECREMENT:
        secondary = core + (program_counter + n) % LIB94_CORE_SIZE;
        secondary->bnumber = (secondary->bnumber + LIB94_CORE_SIZE - 1) % LIB94_CORE_SIZE;
        ptr = (n + secondary->bnumber) % LIB94_CORE_SIZE;
        instr = core[(program_counter + ptr) % LIB94_CORE_SIZE];
        add_read_instruction(secondary);
        add_written_instruction(secondary);
        break;

      case A_INCREMENT:
        secondary = core + (program_counter + n) % LIB94_CORE_SIZE;
        ptr = (n + secondary->anumber) % LIB94_CORE_SIZE;
        instr = core[(program_counter + ptr) % LIB94_CORE_SIZE];
        secondary->anumber = (secondary->anumber + 1) % LIB94_CORE_SIZE;
        add_read_instruction(secondary);
        add_written_instruction(secondary);
        break;

      case B_INCREMENT:
        secondary = core + (program_counter + n) % LIB94_CORE_SIZE;
        ptr = (n + secondary->bnumber) % LIB94_CORE_SIZE;
        instr = core[(program_counter + ptr) % LIB94_CORE_SIZE];
        secondary->bnumber = (secondary->bnumber + 1) % LIB94_CORE_SIZE;
        add_read_instruction(secondary);
        add_written_instruction(secondary);
        break;

    }
  }

  template<opcode op, modifier mod, bool uas>
  bool execute();

  #define EXECUTOR_DECL_LINE(mod, uas) \
    extern template bool execute<DAT, mod, uas>(); extern template bool execute<MOV, mod, uas>(); \
    extern template bool execute<ADD, mod, uas>(); extern template bool execute<SUB, mod, uas>(); \
    extern template bool execute<MUL, mod, uas>(); extern template bool execute<DIV, mod, uas>(); \
    extern template bool execute<MOD, mod, uas>(); extern template bool execute<JMP, mod, uas>(); \
    extern template bool execute<JMZ, mod, uas>(); extern template bool execute<JMN, mod, uas>(); \
    extern template bool execute<DJN, mod, uas>(); extern template bool execute<SEQ, mod, uas>(); \
    extern template bool execute<SNE, mod, uas>(); extern template bool execute<SLT, mod, uas>(); \
    extern template bool execute<SPL, mod, uas>(); extern template bool execute<NOP, mod, uas>();

  #define EXECUTOR_DECL(uas) \
    EXECUTOR_DECL_LINE(A,  uas) \
    EXECUTOR_DECL_LINE(B,  uas) \
    EXECUTOR_DECL_LINE(AB, uas) \
    EXECUTOR_DECL_LINE(BA, uas) \
    EXECUTOR_DECL_LINE(F,  uas) \
    EXECUTOR_DECL_LINE(X,  uas) \
    EXECUTOR_DECL_LINE(I,  uas)

  EXECUTOR_DECL(true)
  EXECUTOR_DECL(false)

  #define EXECUTOR_REF_LINE(mod) \
    &execute<DAT, mod, uas>, &execute<MOV, mod, uas>, \
    &execute<ADD, mod, uas>, &execute<SUB, mod, uas>, \
    &execute<MUL, mod, uas>, &execute<DIV, mod, uas>, \
    &execute<MOD, mod, uas>, &execute<JMP, mod, uas>, \
    &execute<JMZ, mod, uas>, &execute<JMN, mod, uas>, \
    &execute<DJN, mod, uas>, &execute<SEQ, mod, uas>, \
    &execute<SNE, mod, uas>, &execute<SLT, mod, uas>, \
    &execute<SPL, mod, uas>, &execute<NOP, mod, uas>

  template <bool uas>
  static const std::function<bool ()> executors[] = {
    EXECUTOR_REF_LINE(A),  EXECUTOR_REF_LINE(B),
    EXECUTOR_REF_LINE(AB), EXECUTOR_REF_LINE(BA),
    EXECUTOR_REF_LINE(F),  EXECUTOR_REF_LINE(X),
    EXECUTOR_REF_LINE(I)
  };

  static warrior_info *this_warrior;

  void enqueue_process(number_t pc) {
    this_warrior->processes.push_back(pc);
  }

  bool at_process_limit() {
    return this_warrior->processes.size() + 1 == LIB94_MAX_PROCESSES;
  }

  template <bool update_address_sets>
  const warrior *single_step() {
    this_warrior = alive_warriors.front();
    alive_warriors.pop_front();

    program_counter = this_warrior->processes.front();
    this_warrior->processes.pop_front();

    if constexpr (update_address_sets)
      executed_addresses.insert(program_counter);
    instruction_register = core[program_counter];

    evaluate_operand(instruction_register.amode, instruction_register.anumber, a_pointer, a_instruction);
    evaluate_operand(instruction_register.bmode, instruction_register.bnumber, b_pointer, b_instruction);

    a_instruction_writable = core + (program_counter + a_pointer) % LIB94_CORE_SIZE;
    b_instruction_writable = core + (program_counter + b_pointer) % LIB94_CORE_SIZE;

    bool should_endeque = executors<update_address_sets>[instruction_register.op + instruction_register.mod * 16]();

    if (should_endeque)
      this_warrior->processes.push_back((program_counter + 1) % LIB94_CORE_SIZE);

    else if (this_warrior->processes.size() == 0)
      return this_warrior->the_warrior;

    alive_warriors.push_back(this_warrior);
    return 0;
  }

  template const warrior *single_step<true>();
  template const warrior *single_step<false>();

  const warrior *do_round(const instruction &background, const warrior *const *warriors, size_t count, const number_t *offsets, bool shuffle, long steps_to_tie) {

    clear_core(background);
    assert(init_round(warriors, count, offsets, shuffle));
    size_t warriors_left = count;

    for (long i = 0; i < steps_to_tie; ++i)
      if (single_step<false>() != 0) {
        --warriors_left;
        if (warriors_left == 1)
          return get_next_warrior();
      }

    return 0;

  }

  template <bool print_progress>
  void tabulate_mpi(const instruction &background, const warrior *w1, const warrior *w2, long steps_to_tie, int &w1_wins_out, int &w2_wins_out, int &rounds_out, MPI_Comm communicator) {

    assert(w1->instructions.size() + w2->instructions.size() <= LIB94_CORE_SIZE);

    std::vector<number_t> w2_offsets;
    for (number_t i = w1->instructions.size(); i + w2->instructions.size() <= LIB94_CORE_SIZE; ++i)
      w2_offsets.push_back(i);

    int comm_size;
    int comm_rank;

    MPI_Comm_size(communicator, &comm_size);
    MPI_Comm_rank(communicator, &comm_rank);

    if constexpr (print_progress)
      if (comm_rank == 0) {
        for (int i = 0; i < comm_size; ++i)
          std::cout << '\n';
        std::cout << std::flush;
      }

    int actual_count = w2_offsets.size();
    int per_process = (actual_count - 1) / comm_size + 1;

    w2_offsets.resize(per_process * comm_size);
    for (int i = actual_count; i < per_process * comm_size; ++i)
      w2_offsets[i] = LIB94_CORE_SIZE + 1;

    std::shuffle(w2_offsets.begin(), w2_offsets.end(), prng);

    number_t *recv_buffer = new number_t[per_process];
    MPI_Scatter(w2_offsets.data(), per_process, LIB94_MPI_NUMBER_T, recv_buffer, per_process, LIB94_MPI_NUMBER_T, 0, communicator);

    int wins[2] = {0, 0};

    const warrior *w1_first_warriors[2] = {w1, w2};
    number_t w1_first_offsets[2] = {0, 0};

    const warrior *w2_first_warriors[2] = {w2, w1};
    number_t w2_first_offsets[2] = {0, 0};

    int last_percent = -1;

    for (int i = 0; i < per_process; ++i)
      if (recv_buffer[i] != LIB94_CORE_SIZE + 1) {

        if constexpr (print_progress) {
          int percent = i * 100 / per_process;
          if (percent != last_percent) {
            std::cout << "\x1b[" << comm_size - comm_rank << "A\x1b[3G";
            std::cout << "worker " << comm_rank + 1 << ": " << percent << "%\x1b[0K";
            std::cout << "\x1b[" << comm_size - comm_rank << "B\x1b[1G";
            std::cout << std::flush;
          }
        }

        w1_first_offsets[1] = recv_buffer[i];
        const warrior *w = do_round(background, w1_first_warriors, 2, w1_first_offsets, false, steps_to_tie);
        if (w == w1)
          ++wins[0];
        else if (w == w2)
          ++wins[1];

        w2_first_offsets[0] = recv_buffer[i];
        w = do_round(background, w2_first_warriors, 2, w2_first_offsets, false, steps_to_tie);
        if (w == w1)
          ++wins[0];
        else if (w == w2)
          ++wins[1];

      }


    if constexpr (print_progress) {
      std::cout << "\x1b[" << comm_size - comm_rank << "A\x1b[3G";
      std::cout << "worker " << comm_rank + 1 << ": complete\x1b[0K";
      std::cout << "\x1b[" << comm_size - comm_rank << "B\x1b[1G";
      std::cout << std::flush;
    }

    int win_sums[2];

    MPI_Allreduce(wins, win_sums, 2, MPI_INT, MPI_SUM, communicator);
    w1_wins_out = win_sums[0];
    w2_wins_out = win_sums[1];
    rounds_out = actual_count * 2;

  }

  template void tabulate_mpi<true>(const instruction &background, const warrior *w1, const warrior *w2, long steps_to_tie, int &w1_wins_out, int &w2_wins_out, int &rounds_out, MPI_Comm communicator);
  template void tabulate_mpi<false>(const instruction &background, const warrior *w1, const warrior *w2, long steps_to_tie, int &w1_wins_out, int &w2_wins_out, int &rounds_out, MPI_Comm communicator);

}
